from math import *
import numpy as np
from numpy import linalg as LA
import matplotlib.pyplot as plt


# Methode de Jacobi
def jacobi(A,b,x0,eps,maxiter) :
    x = x0; xt = x; XX = [];
    D = A * np.eye(n);
    iter = 0; err = 1;
    while err>eps and iter<maxiter :
        y = A@x - b;
        xt = x - LA.solve(D,y);
        iter += 1; 
        err = LA.norm(x-xt);
        x = xt;
        XX.append(x);
    return XX


# Construction des deux matrices tests
n = 15;
A = 2*np.eye(n,n) - np.diag([1 for i in range(n-1)],1) - np.diag([1 for i in range(n-1)],-1) 
B = 4*np.eye(n,n) + np.diag([1 for i in range(n-1)],1) + np.diag([1 for i in range(n-1)],-1) 

# On calcule un second membre tel que la solution du système soit un vecteur de 1. 
xex = np.ones((n,1)); b = A@xex; c = B@xex;
x0 = np.zeros((n,1));

# Resolution :
x = jacobi(A,b,x0,1e-12,1000);
x=np.array(x)
y = jacobi(B,c,x0,1e-12,1000);
y=np.array(y)

# Calcul de l'erreur et illustration de la convergence
[pA,qA]=np.shape(x)[0:2]; errA=[];
for i in range(pA) :
  errA.append(LA.norm(xex-x[i,:,0]));
errA=np.array(errA);
[pB,qB]=np.shape(y)[0:2]; errB=[];
for i in range(pB) :
  errB.append(LA.norm(xex-y[i,:,0]));
errB=np.array(errB);

plt.clf()  
plt.plot(range(pA),np.log(errA), label="matrice laplacien")
plt.plot(range(pB), np.log(errB), label="matrice spline") 
plt.title("Convergence methode de Jacobi - échelle logarithmique pour l'erreur")
plt.xlabel('iterations')
plt.ylabel('Erreur')
plt.legend()
plt.show ()

#Taux de convergence - matrice A
print('Taux de convergence estime (matrice A) : ')
[a1,a2] = np.polyfit (range(pA), np.log (errA) ,1)
print(exp(a1))
print('Rayon spectral de la matrice des iterations (laplacien) : ')
D = A*np.eye(n); 
print( max(abs(LA.eigvals(np.eye(n,n)-LA.inv(D)@A)))) 
# rayon spectral de M^{-1}N

#Taux de convergence - matrice B
print('Taux de convergence estime (matrice B) : ')
[a1,a2] = np.polyfit (range(pB), np.log (errB) ,1)
print(exp(a1))
print('Rayon spectral de la matrice des iterations (splnes) : ')
D = B*np.eye(n); 
print( max(abs(LA.eigvals(np.eye(n,n)-LA.inv(D)@A)))) 
# rayon spectral de M^{-1}N



# Methode de Jacobi relaxee
def jacobi_relax (A,b,omega,x0,eps,maxiter) :
    x = x0; xt = x; XX = [];
    D = A * np.eye(n);
    M = D/omega ; N = M-A;
    iter = 0; err = 1;
    while err>eps and iter<maxiter :
        y = N@x + b;
        xt = LA.solve(M,y);
        iter += 1; 
        err = LA.norm(x-xt);
        x = xt;
        XX.append(x);
    XX=np.array(XX)
    return XX



# Test Jacobi relaxee matrices A et B 
omega=0.6
x = jacobi_relax(A,b,omega,x0,1e-6,1000);
[pA2,qA2]=np.shape(x)[0:2]; errA2=[];
for i in range(pA2) :
  errA2.append(LA.norm(xex-x[i,:,:]));
errA2=np.array(errA2);
y = jacobi_relax(B,c,omega,x0,1e-6,1000);
[pB2,qB2]=np.shape(y)[0:2]; errB2=[];
for i in range(pB2) :
  errB2.append(LA.norm(xex-y[i,:,:]));
errB2=np.array(errB2);

plt.clf()  
plt.plot(range(pA2),np.log(errA2), label="matrice laplacien")
plt.plot(range(pB2),np.log(errB2), label="matrice splines")
plt.title("Convergence methode de Jacobi relaxée - omega ="+str(omega))
plt.xlabel('iterations')
plt.ylabel('log(Erreur)')
plt.legend()
plt.show ()

# Etude du taux de convergence selon la valeur de omega, pour A
omegas=np.arange(0,1.2,0.1);
omegas=omegas[1:];
taux=[];
rhospec=[];
for omega in omegas :
    # tirage d'un vecteur solution aléatoire
    xex = np.random.rand(n);
    xex = xex/LA.norm(xex);
    b=A*xex;
    x = jacobi_relax(A,b,omega,x0,1e-12,200);
    [p,q] = np.shape(x)[0:2];
    err = [];
    for i in range(p) :
        err.append(LA.norm(xex-x[i,:,:]));
    err=np.array(err)
    plt.plot(range(p),np.log(err), label="omega="+str(omega));
    plt.title('Courbe de convergence (echelle log en ordonnee)')
    plt.xlabel('omega')
    plt.ylabel('log(Erreur)')
    plt.legend()
    plt.show()
    [a1,a2] = np.polyfit(range(p),np.log(err),1);
    taux.append(exp(a1));
    rhospec.append(max(abs(LA.eigvals(np.eye(n)-omega*LA.inv(A*np.eye(n))@A))));



plt.clf();
plt.plot(omegas,np.transpose(taux),label='taux')
plt.plot(omegas,np.transpose(rhospec),label='rayon spectral')
plt.title('Taux de convergence en fonction de omega - matrice du laplacien')
plt.xlabel('omega')
plt.ylabel('Taux de convergence') 
plt.legend()
plt.show()


# Etude du taux de convergence selon la valeur de omega, matrice B
omegas=np.arange(0,1.2,0.1);
omegas=omegas[1:];
taux=[];
rhospec=[];
epsilon=1e-12;
for omega in omegas :
    y = jacobi_relax(B,c,omega,x0,1e-12,200);
    [p,q] = np.shape(y)[0:2];
    err = [];
    for i in range(p) :
        err.append(LA.norm(xex-y[i,:,:]));
    err=np.array(err)
    plt.plot(range(p),np.log(err), label="omega="+str(omega));
    plt.title('Courbe de convergence (echelle log en ordonnee)')
    plt.xlabel('omega')
    plt.ylabel('Erreur')
    plt.legend()
    plt.show()
    [a1,a2] = np.polyfit(range(p),np.log(err),1);
    taux.append(exp(a1));
    rhospec.append(max(abs(LA.eigvals(np.eye(n)-omega*LA.inv(A*np.eye(n))@A))));



plt.clf();
plt.plot(omegas,np.transpose(taux),label='taux')
plt.plot(omegas,np.transpose(rhospec),label='rayon spectral')
plt.title('Taux de convergence en fonction de omega - matrice spline')
plt.xlabel('omega')
plt.ylabel('Taux de convergence')
plt.legend()
plt.show()

# Fonction Chebyshev-Jacobi
def chebyshev_jacobi(A,b,x0,eps,maxiter,m1,m2):
    D = A*np.eye(n)
    k = (1-m1)/(1-m2)
    iter = 0; sig = 2; x = x0;
    res = A@x - b;
    XX = [x];
    SIGMA = [sig] 
    iter += 1
    xt = x;
    sig = 4/(4-((k-1)/(k+1))**2*sig);
    x = xt-2/(2-m1-m2)*LA.solve(D,res);
    res = A@x - b;
    XX.append(x);
    SIGMA.append(sig)
    iter += 1 
    err = 1
    while err>eps and iter < maxiter :
        xtt = xt
        xt = x
        sig = 4/(4-((k-1)/(k+1))**2*sig);
        x = sig*(xt-2/(2-m1-m2)*LA.solve(D,res))+(1-sig)*xtt ;
        err = LA.norm(x-xt)
        res = A@x - b;
        XX.append(x);
        SIGMA.append(sig)
        iter += 1 
    XX = np.array(XX)
    SIGMA = np.array(SIGMA)
    return [XX,SIGMA]


# Test Chebyshev-Jacobi matrice A

m1=-0.9;
m2=0.9;
[xC,SIGMA] = chebyshev_jacobi(A,b,x0,1e-15,800,m1,m2);
[p,q] = np.shape(xC)[0:2];
err=[];
for i in range(p) :
  err.append(np.linalg.norm(xex-xC[i,:,:],inf));

# Tracé des courbes de convergence - matrice du laplacien
plt.clf
plt.plot(range(p),np.log(err), label="Chebyshev-Jacobi")
plt.plot(range(pA),np.log(errA), label="Jacobi")
plt.plot(range(pA2),np.log(errA2), label="Jacobi relaxee")
[a1,a2] = np.polyfit(range(p),np.log(err),1)
plt.title('Courbe de convergence (matrice du laplacien)')
plt.xlabel('Itération')
plt.ylabel('log(Erreur)')
plt.legend()
plt.show()
